<?php
/** vim: set tabstop=4 softtabstop=4 shiftwidth=4 textwidth=80 smarttab expandtab: **/
/** coding: utf-8: **/
/*
 * Copyright (C) 2012  Sangoma Technologies Corp.
 * All Rights Reserved.
 *
 * Author(s)
 * your name <your_name@sangoma.com>
 *
 * This code is Sangoma Technologies Confidential Property.
 * Use of and access to this code is covered by a previously executed
 * non-disclosure agreement between Sangoma Technologies and the Recipient.
 * This code is being supplied for evaluation purposes only and is not to be
 * used for any other purpose.
*/
/*
 * OS class
*/
require_once ('application/helpers/safe_helper.php');
safe_require_class('safe_auditor_class');
class Safe_os_wrapper extends Safe_object_class
{
    public function __construct()
    {
        parent::__construct('', 'os');
    }
    public function execute($cmd, &$result = null, $escape = true, $superuser = false, &$return_var=null, $log_error=true)
    {
        // PHP handle the os specific exec stuff
        // don't need to delegate to os specific class
        $return_var = 0;
        if ($superuser) $result = safe_sudo_exec($cmd, '', '', $return_var, $escape);
        else $result = safe_exec($cmd, '', '', $return_var, $escape);
        if (0 !== $return_var && $log_error){
            error_log('Exec <' . $cmd . '> failed=' . $return_var . ' output=' . print_r($result, true));
        }
        return ($return_var === 0);
    }
    public function service_ctl($service_name, $action, &$result = null, $process_name = null)
    {
        return false;
    }
    public function mkdir($pathname, $chown_recursive=true)
    {
        if (!file_exists($pathname)){
            Safe_sudo_exec('/bin/mkdir -p ' . $pathname);
            clearstatcache(true, $pathname);
        }
        if($chown_recursive == true){
            Safe_sudo_exec('/bin/chown -R webconfig ' . $pathname);
        }else{
            // Check if root is not file owner (test with 0 being root uid
            if(!fileowner($pathname)) {
                Safe_sudo_exec('/bin/chown webconfig ' . $pathname);
            }
        }
        return file_exists($pathname);
    }
    public function rmdir($pathname)
    {
        if (file_exists($pathname)){
            Safe_sudo_exec('/bin/rm -rf ' . $pathname);
            clearstatcache(true, $pathname);
        }
        return (!file_exists($pathname));
    }
    public function delete_file($file_name)
    {
        if (file_exists($file_name)){
            Safe_sudo_exec('/bin/rm -fr ' . $file_name);
            clearstatcache(true, $file_name);
        }
        return (!file_exists($file_name));
    }
    public function move_file($src_file, $dst_file)
    {
        if (file_exists($src_file)){
            Safe_sudo_exec('/bin/chown webconfig ' . $src_file);
            Safe_sudo_exec('/bin/mv -f ' . $src_file . ' ' . $dst_file);
            clearstatcache(true, $src_file);
            clearstatcache(true, $dst_file);
        }
        return (file_exists($dst_file));
    }
    public function copy_file($src_file, $dst_file)
    {
        if (file_exists($src_file)){
            Safe_sudo_exec('/bin/cp -f ' . $src_file . ' ' . $dst_file);
            clearstatcache(true, $src_file);
            clearstatcache(true, $dst_file);
        }
        return (file_exists($dst_file));
    }

    private function _recursiveGlob($dir, $ext) {
        $globFiles = array();
        if(isset($dir) && strlen($dir) && '/' !== $dir){
            $globFiles = glob("$dir/$ext", GLOB_BRACE);
            $globDirs  = glob("$dir/*", GLOB_ONLYDIR);
            foreach ($globDirs as $rdir) {
                $files = $this->_recursiveGlob($rdir, $ext);
                $globFiles = array_merge($globFiles, $files);
            }
        }

        return $globFiles;
    }
    
    public function search_files($pattern, $recursive=false, $sort=true)
    {
        if($recursive){
            $dir = dirname($pattern);
            $base = basename($pattern);
            $files = $this->_recursiveGlob($dir, $base);
        }else{
            $files = glob($pattern, GLOB_BRACE);
            if (false == $files) $files = array();
        }
        if($sort){
            array_multisort(array_map('filemtime', $files), SORT_NUMERIC, SORT_DESC, $files);
        }
        return $files;
    }
    public function check_file_exists($file_name, $recursive=false)
    {
        if($recursive){
            $dir = dirname($file_name);
            $base = basename($file_name);
            $files = $this->_recursiveGlob($dir, $base);
            return (0 != count($files));
        }else{
            return (file_exists($file_name));
        }
    }
    public function find_file($dir, $file_name, $recursive=false)
    {
        if($recursive){
            $files = $this->_recursiveGlob($dir, $file_name);
        }else{
            $files = glob("$dir/$file_name", GLOB_BRACE);
        }
        if(isset($files[0])){
            return $files[0];
        }
        return false;
    }
    public function stat_file($file, &$stats)
    {
        clearstatcache(true, $file);
        $stats = stat($file);
        return ($stats);
    }
    /**
     * Creates a symbolic link. Must use absolute paths
     * @param string  $original
     *                           			Original file location
     * @param string  $target  
     *                           			Location where symbolic link
     * @return boolean
     */
    public function symlink_file($original, $target)
    {
        if (file_exists($original)) {
            // Ensure destination directory exists
            $this->mkdir(dirname($target));
            Safe_sudo_exec('/bin/chown webconfig ' . $original);
            Safe_sudo_exec('/bin/rm -f ' . $target);
            Safe_sudo_exec('/bin/ln -sf ' . $original . ' ' . $target);
            Safe_sudo_exec('/bin/chown webconfig ' . $target);
            return true;
        }
        return false;
    }
    /**
     * Unlinks a symbolic link
     * @param string  $file
     *                       			Absolute path to symobolic link
     * @return boolean
     */
    public function unlink_file($file)
    {
        if (file_exists($file)){
            Safe_sudo_exec('/bin/rm -f ' . $file);
            clearstatcache(true, $file);
        }
        return true;
    }
    public function write_file($file_name, $buffer, $mode = 'w')
    {
        // Ensure directory exist
        if ($this->mkdir(dirname($file_name))) {
            if (file_exists($file_name)) Safe_sudo_exec('/bin/chown webconfig ' . $file_name);
            if ($file = fopen($file_name, $mode)) {
                if (is_array($buffer)) {
                    foreach ($buffer as $line) fputs($file, $line . "\n");
                } else {
                    fwrite($file, $buffer);
                }
                fclose($file);
                return true;
            }
        }
        return false;
    }
    /**
     * Returns the output of a file
     *
     * @param $file_name = file to open and read
     * @param $buffer    = variable to store file contents in
     * @param mode       = mode to open file in (r, rw, a)
     * @param $as_array  = true to store as array, false to store as string
     *                      
     * @return true       or false
     */
    public function read_file($file_name, &$buffer, $mode = 'r', $as_array = false)
    {
        // TODO(wadam) - use local read file from php
        if (file_exists($file_name)) {
            if ($file = fopen($file_name, $mode)) {
                $buffer = '';
                if ($as_array) {
                    $buffer = array();
                    while ($line = fgets($file)) $buffer[] = $line;
                } else {
                    while (!feof($file)) {
                        $buffer.= fread($file, filesize($file_name));
                    }
                }
                fclose($file);
                return true;
            } else {
                // Fail to open, probably user right issue use cat
                $buffer = Safe_sudo_exec('/bin/cat ' . $file_name, '', '', $rc);
                return (0 === $rc);
            }
        }
        //error_log('Fail to open file' . $file_name);
        return false;
    }
    /**
     * chmod file
     * @param string  $file_name = file to change
     * @param string  $mod  = new file mode
     * @param string  $include_sub  = includ subdir
     *                       			
     * @return boolean
     */
    public function chmod_file($file_name, $mod)
    {
        if (file_exists($file_name)){
            Safe_sudo_exec('/bin/chmod ' .  $mod . ' ' . $file_name,'', '', $rc);
            return (0 === $rc);
        }
        return false;
    }
    /**
     * chown file
     * @param string  $file_name = file to change
     * @param string  $owner  = new file mode
     * @param string  $include_sub  = includ subdir
     *
     * @return boolean
     */
    public function chown_file($file_name, $owner)
    {
        if (file_exists($file_name)){
            Safe_sudo_exec('/bin/chown ' . $owner . ' ' . $file_name,'', '', $rc);
            return (0 === $rc);
        }
        return false;
    }
    /**
     * SED  file
     * @param string  $file_name = file to change
     * @param string  $data  =is an array of SED commands
     *
     * @return boolean
     */
    public function sed_file($file_name, $data = null)
    {
        if (file_exists($file_name)){
            foreach($data as $argument){
                Safe_sudo_exec('/bin/sed -i \'' . $argument . '\' ' . $file_name, '', '', $rc, false);
            }
            return true;
        }
        return false;
    }
    /**
     * Test if $port is a valid port can LISTEN
     * @param string $ip
     * @param string $port
     * @param string $type = (UDP,TCP,ALL)
     * @param string $program_name
     */
    public function valid_ip_port($ip, $port, $type_str = 'TCP', $program_name = '', &$user_prog='') {
        $ip     = strtoupper(trim($ip));
        $type_str = strtoupper(str_ireplace( array ('TLS','ALL'), array('TCP','TCPUDP'), $type_str));
        $param = '-lpn';
        if(strpos($type_str,'TCP')!==false){
            $param = $param .'t';
        }
        if(strpos($type_str,'UDP')!==false){
            $param = $param .'u';
        }
        $output = array();
        $cmd = "netstat {$param} | grep ':$port '";
        // Loop X times on port check as it not always reliable
        // See bug#8270
        $port_used_count = 0;
        $loop_cnt = 0;
        $max_loop = 5;
        // Default is used
        $port_used = true;
        do{
            // Default is used
            $port_used = true;
            // Check for port usage
            $output = safe_sudo_exec($cmd, '', '', $return_var, false);
            if($output != null ){
                // Go matches, loop around result
                foreach($output as $matched_string){
                    if($ip != 'ALL'){
                        //check if ip match or not
                        if(strpos($matched_string,':::'.$port) !== false || strpos($matched_string,'0.0.0.0:'.$port) !== false){
                        }else{
                            if(strpos($matched_string,$ip.':'.$port) === false){                                                                                                             
                                $port_used = false;
                            }
                        }
                    }
                    $user_prog = substr(strrchr($matched_string, '/' ), 1 );
                    if($port_used == true && $program_name != '' 
                        && ($user_prog == $program_name || $user_prog== 'sh') )
                    {
                        $port_used = false;
                    }
                    if($port_used) {
                        error_log('valid_ip_port('.$ip.','.$port.','.$type_str.','.$program_name.') used by <'.$matched_string.'>');
                        $port_used_count++;
                        break;
                    }
                }
            }else{
                $port_used = false;
            }
        } while($loop_cnt++<$max_loop);

        // Check if port marked as used 'max_loop' times
        if($port_used && $port_used_count<$max_loop){
            error_log('valid_ip_port('.$ip.','.$port.','.$type_str.','.$program_name.') marked as used only '.$port_used_count.' - false positive');
            $port_used = false;
        }

        return !$port_used;
    }
    private $_hostname=null;
    public function hostname()
    {
        if(!$this->_hostname){
            $hostname[] = "localhost";
            $this->execute('hostname', $hostname);
            $this->_hostname = $hostname[0];
        }
        return $this->_hostname;
    }
    public function statistics()
    {
        return array();
    }
    public function version()
    {
        return false;
    }
    public function can_install_data($info, &$reason)
    {
        return true;
    }
}
class Safe_centos_wrapper extends Safe_os_wrapper
{
    public function service_ctl($service_name, $action, &$result = null, $process_name = null)
    {
        $return_var = 0;
        if (isset($process_name)) $service_name = $process_name;
        switch ($action) {
        case 'status':
            safe_sudo_exec('service', $service_name . ' ' . $action, '', $return_var, false);
            if (!isset($result)) break;

        case 'stats':
            /*
            $result['process'] = '';
            $result['cpu']['time'] = '';
            $result['cpu']['percent'] = '';
            $result['mem']['percent'] = '';
            $result['mem']['kb'] = '';
             */
            $output = safe_exec('ps --no-headers -eo comm,etime,%cpu,%mem,rss| grep -e "^' . $service_name . '[ \\t]\\+[0-9]"', '', '', $dummy, false);
            if(!$dummy){
                if(empty($output)){
                    $return_var = - 1;
                }else{
                    foreach($output as $line){
                        $line = str_replace('  ', ' ', $line);
                        $stats = explode(' ', trim(str_replace($service_name, ' ', $line)));
                        $result['process'] = $service_name;
                        $result['cpu']['time'] = $stats[0];
                        $result['cpu']['percent'] += $stats[1];
                        $result['mem']['percent'] += $stats[2];
                        $result['mem']['kb'] += $stats[3];
                    }
                }
                // Adjust cpu % to cpu_count
                $result['cpu']['percent'] = $result['cpu']['percent'] / $this->cpu_count();
            }
            break;

        default:
            // Log only default as we don't want to see service xxx status in log
            error_log('service_ctl('.$service_name.', '.$action.') = '.$return_var);
            $result = safe_sudo_exec('/sbin/service ' . $service_name . ' ' . $action, '', '', $return_var, false);
            break;
        }
        return $return_var;
    }
    public function set_service_startup($service_name, $is_auto, $process_name=null)
    {
        if (isset($process_name)) $service_name = $process_name;
        $rc = null;
        if ($is_auto) {
            safe_sudo_exec('chkconfig --add ' . $service_name, $rc);
            $output = safe_sudo_exec('chkconfig ' . $service_name . ' on', $rc);
        } else {
            $output = safe_sudo_exec('chkconfig ' . $service_name . ' off', $rc);
        }
        return (0 == $rc);
    }
    public function is_service_auto($service_name, $process_name=null)
    {
        if (isset($process_name)) $service_name = $process_name;
        $output = array();
        $output = safe_sudo_exec('chkconfig --list ' . $service_name . ' | grep :on', '', '', $rc, false);
        return (empty($output)) ? false : true;
    }
    public function kill($service_name, $signal, $process_name = null)
    {
        if (isset($process_name)) $service_name = $process_name;
        $rc = null;
        $result = safe_sudo_exec('/usr/bin/pkill', '-' . $signal . ' ' . $service_name, '', $rc, false);
        return (0 == $rc);
    }
    public function statistics()
    {
        $stats = array();

        // Get /proc/meminfo
        $output = array();
        $rc = $this->execute('nohup cat /proc/meminfo', $output);
        if($rc){
            /* Sample output
                root@mtl-nsc-devel-will webconfig
                cat /proc/meminfo
                MemTotal:         513340 kB
                MemFree:           42120 kB
                Buffers:          108340 kB
                Cached:           130036 kB
                ...
             */
            $memtotal = explode(" ", preg_replace('/\s+/', ' ', $output[0]));
            $memfree = explode(" ", preg_replace('/\s+/', ' ', $output[1]));
            $membuffers = explode(" ", preg_replace('/\s+/', ' ', $output[2]));
            $memcached = explode(" ", preg_replace('/\s+/', ' ', $output[3]));

            $memtotal = intval(trim($memtotal[1]));
            $memfree = intval(trim($memfree[1]));
            $membuffers = intval(trim($membuffers[1]));
            $memcached = intval(trim($memcached[1]));
            // In kb
            $stats['memory']['kb']['total'] = $memtotal;
            $stats['memory']['kb']['used'] = $stats['memory']['kb']['total'] - $memfree - $membuffers - $memcached;
            $stats['memory']['kb']['free'] = $memfree + $membuffers + $memcached;
            // In percent
            $stats['memory']['percent']['free'] = floatval(($stats['memory']['kb']['free'] / $stats['memory']['kb']['total'])*100.0);
            $stats['memory']['percent']['used'] = 100.0 - $stats['memory']['percent']['free'];
        }
        // Get /proc/meminfo
        $output = array();
        $rc = $this->execute('nohup cat /proc/stat', $output);
        if($rc){
            /* Sample output
            root@mtl-nsc-devel-will webconfig
            cat /proc/stat
            cpu  1518057 3072 1000613 154507988 64384 0 336660 0 0 0
            ...
             */
            $cpu_info = explode(" ", preg_replace('/\s+/', ' ', $output[0]));
            unset($cpu_info[0]);
            // Calculate cpu total
            $cpu_total = 0;
            foreach($cpu_info as $k => $v){
                $v = intval($v);
                $cpu_info[$k] = $v;
                $cpu_total += $v;
            }
            // Various cpu values: http://linux.die.net/man/5/proc
            // Latch absolute values
            $curr_stats['total'] = $cpu_total;
            // user = user + nice
            $curr_stats['user'] = $cpu_info[0] + $cpu_info[1];
            $curr_stats['system'] = $cpu_info[2];
            $curr_stats['wait'] = $cpu_info[3];
            $curr_stats['idle'] = $cpu_info[4];
            // Get previous counters from $_SESSION
            $prev_stats = array();
            if($_SESSION['system']['stats']){
                $prev_stats = $_SESSION['system']['stats'];
            }
            // Store stats for next round
            session_start();
            $_SESSION['system']['stats'] = $curr_stats;
            session_write_close();
            // Compute difference since last run
            $_total = $curr_stats['total'] - $prev_stats['total'];
            foreach($curr_stats as $k=>$v){
                $stats['cpu']['percent'][$k] = floatval((($curr_stats[$k]-$prev_stats[$k])/$_total)*100.0);
            }
        }

        return $stats;
    }
    public function cpu_count()
    {
        if($_SESSION['system']['cpu']['count']){
            return $_SESSION['system']['cpu']['count'];
        }
        // Default is 1 CPU at least ;)
        $cpu_count = 1;
        $rc = $this->execute('/usr/bin/nproc', $output);
        if($rc){
            $cpu_count = intval($output[0]);
        }
        session_start();
        $_SESSION['system']['cpu']['count'] = $cpu_count;
        session_write_close();
        return $cpu_count;
    }
    public function version()
    {
        $uname_field = array('kernel-name', 'nodename',  'kernel-release', 'kernel-version', 'machine', 'processor', 'hardware-platform', 'operating-system');
        $result = array();
        foreach($uname_field as $item){
            $output = array();
            $rc = $this->execute('nohup uname --' . $item, $output);
            if($rc){
                $result[$item] = $output[0];
            }
        }
        return $result;
    }
    /**
     * @brief Check if update package can be install
     *
     * @param[in out] $reason
     *
     * @return true/false
     */
    public function can_install_data($info, &$reason)
    {
        if('update' == $info['type']){
            $system_info = $this->version();
            if ($system_info['hardware-platform'] != $info['architecture']) {
                $reason[] = 'Cannot install ' . $info['type'] . ' for ' . $info['architecture'] .' on a ' . $system_info['hardware-platform'] . ' platform.';
                return false;
            }
        }
        return parent::can_install_data($info, $reason);
    }
}
class Safe_debian_wrapper extends Safe_os_wrapper
{
    public function service_ctl($service_name, $action, &$result = null)
    {
        $return_var = 0;
        $result = safe_exec('/etc/init.d/' . $service_name, $action, '', $return_var);
        return $return_var;
    }
}
/* End of file safe_os_class.php */
